#   Copyright: Australian Bureau of Meteorology [2011]
#   Authors:
#       [Tennessee Leeuwenburg <t.leeuwenburg@bom.gov.au>]
#       [Edward Schofield <ed@pythoncharmers.com>]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# Core imports
import datetime
import cProfile
import pstats
import os
import tempfile
import traceback
import urllib
import csv

# Non-core imports
import decorator

# Package imports
# NONE

# To provide basic out-of-the-box functionality, a temp file is created to store 
# the profile stats. However, to make best use of run charting, a 
tempdir = tempfile.gettempdir()
prefix = tempfile.gettempprefix()
TEMP_LOGFILE_NAME = tempdir + '/' + prefix + '.pstats'
LOGFILE_NAME = TEMP_LOGFILE_NAME
STOPPED = False
VERSION_TAG = None

# Initialise a single module-level profiler object to collect data
profiler = cProfile.Profile()

#TODOS
# Need to implement or remove the max_time_seconds argument 


def set_log_file_name(filename):
    '''
    Use this to set the logfile name to use for storing profile stats
    '''
    global LOGFILE_NAME
    LOGFILE_NAME = filename
    
    global LOGFILE_TEMP_NAME
    LOGFILE_TEMP_NAME = None
    
    global STOPPED
    STOPPED = False
    
def set_version_tag(tag):
    '''
    In addition to the datetime-based filename structure on disk, 
    benchmarker.py will allow each benchmark run to be tagged with some
    kind of version identifier. Uses for this could include:
        - Revision control version identifier
        - Experimental run identifier
    
    This may need to get fleshed out into a more substantive metadata
    format at some point, but starting with a tag seems okay.
    '''
    
    global VERSION_TAG
    safe_tag = urllib.quote(tag)
    VERSION_TAG = safe_tag
    
def set_archiving(top_level_dir_name, 
                  project_name, 
                  environment_name, 
                  executable_name):
    '''
    With this mechanism, benchmarker.py will automatically store an archive of benchmark
    results. It will create a filesystem structure along the following lines:
    
    /top_level_dir_name
       /project_name
           /environment_name
               /executable_name
                   /year-as-int (01, 02.. 12)
                      /month-as-int (01, 02.. 12)
                         /day-of-month-as-int (01, 02.. 31)
                            YYYY_MM_DD_HH_MM_SS.pstats
                            ...
              
    This structure allows daily, weekly and monthly benchmarking to occur fairly naturally and
    about the right kind of complexity for a human to understand as well.
    
    Each time this function is called, the current module-level profiler will be closed (just in case)
    and a new pstats file created based on the current timestamp. Under normal use, this will
    then give one pstats file per test execution run. However, having the ability for this function
    to be called explicitly means that it can be manually controlled. This could make is usable
    in a live-benchmarking environment where the interpreter instance is long-lived for example
    '''
    
    # Using the module-level profiler here
    global profiler
    global LOGFILE_NAME
    global STOPPED
    
    try:
        profiler.dump_stats(LOGFILE_NAME)
    except:
        print "There is every chance this is not a big deal, but please read the traceback to check"
        traceback.print_exc()
    
        
    now = datetime.datetime.utcnow()
    
    now_as_string = now.strftime('%Y_%m_%d_%H_%M')
    year_as_string = now.strftime("%Y")
    month_as_string = now.strftime("%m")
    day_as_string = now.strftime('%d')
    
    directory = '/'.join([top_level_dir_name, 
                          project_name,
                          environment_name,
                          executable_name,
                          year_as_string,                           
                          month_as_string, 
                          day_as_string])
    
    # Make the directory only if needed
    if not os.path.exists(directory):
        os.makedirs(directory)
    
    filename = directory + '/' + now_as_string + ".pstats"
    if VERSION_TAG:
        filename = directory + '/' + VERSION_TAG + '__' + now_as_string + ".pstats"
    
    # It's certainly possible that someone running very quick tests could do this
    # more than once a minute, so handle filename collisions
    i = 1
    while os.path.exists(filename):
        filename = directory + '/' + now_as_string + "_" + repr(i) + ".pstats"
        
        if VERSION_TAG:
            filename = directory + '/' + VERSION_TAG + "__" + now_as_string + "_" + repr(i) + ".pstats"
        i += 1

    LOGFILE_NAME = filename    
    STOPPED = False
    profiler = cProfile.Profile()
    

class benchmark(object):
    '''
    This class is a decorator which can be applied to any function to
    submit it to the profiler. 
    '''
    
    def __init__(self, repeat=100, max_time_seconds=10):
        '''
        Set up the desired number of repeats for this function, plus the 
        maximum allowed number of seconds. A test failure is raised if
        this is exceeded while trying to reach the desires number of repeats
        '''
        
#        print "Initialising a benchmarker" # Can be used for a bit of light logging
        self.repeat = repeat
        self.max_time_seconds = max_time_seconds
        
    def __call__(self, function_to_call):
        '''
        This method is called at the time of decoration (i.e. just after __init__
        and returns the function will should get executed when the inner function is called.
        '''
        
        def wrapped_function_call(*args, **kwargs):
            '''
            This is what will actually get called when the decorated function is called.
            
            The inner function is run a specified number of times to exercise the code for
            the profiler. The return value is whatever the inner function returns the
            last time it is run.
            
            If the inner function raises an exception, it will be caught and stored, and
            the inner function will still be run N times. After that conclusion, the 
            stored exception is re-raised.
            
            @return the results of the inner function
            '''
            
            retval = None
            exception = None
            
            args = args[1:] # Strip off the function name from the args list here
            
            for _ in xrange(self.repeat):
                
                try:
                    retval = profiler.runcall(function_to_call, *args, **kwargs)
                except Exception as e:
                    exception = e
                
            if exception:
                raise exception
            
            return retval

        return decorator.decorator(wrapped_function_call, function_to_call)
            
        

def runcall(function_to_call, *args, **kwargs):
    '''
    Thin wrapper to the profiler module. This hook allows users to use non-decorator
    based benchmarking, but still taking advantage of the automatic stats file management.
    
    i.e., if used in conjunction with appropriate calls to set_archiving(), can do
    benchmarking calls and the stats objects will get automatically archived for later
    visualisation.
    '''
    
    profiler.runcall(function_to_call, *args, **kwargs)
       
def stop_recording():
    '''
    Add a hook to allow the profiling to be explicitly stopped, forcing
    a flush to disk if necessary. Should not be strictly required.
    '''
    
    global STOPPED
    
    if not STOPPED:
        profiler.dump_stats(LOGFILE_NAME)
        STOPPED = True
        
def suspend_recording():
    '''
    A hook to allow the profiling to be explicitly stopped. Stats are
    NOT dumped to disk.
    '''
    
    global STOPPED
    STOPPED = True
        
def restart_recording():
    '''
    This function instructs the module to continue profiling using the current
    Profiler instance. It will clear the STOPPED module variable and delete
    any logfiles created on disk, since the same filename will get written
    to when profiling is later stopped again
    '''
    
    global STOPPED
    
    # No sense restarting if not already stopped!
    if not STOPPED:
        return
    
    # Delete the log file, since it will be re-written later when
    # profiling is fully completed
    os.remove(LOGFILE_NAME)
    STOPPED = False
    
    
def print_performance_report(limit=25):
    '''
    Right now, this is just a worse version of pstats.print_stats(), but
    in the very near future it should be expanded to include relative stats,
    variance between runs etc
    '''

    global STOPPED
    
    # Need to stop recording for a moment to flush all the stats to disk
    if not STOPPED:
        stop_recording()
            
    stats = pstats.Stats(LOGFILE_NAME)
    print stats.total_tt
    
    for index, (key, value) in enumerate(stats.stats.items()):
        
        if index > limit:
            break
        
        filename, count, func_name = key
        ccalls, ncalls, total_time, cumulative_time, callers = value
        print "%s: (ncalls %s), (tot_time %s), (cum_time %s)" % (func_name, 
                                                             ncalls,
                                                             total_time,
                                                             cumulative_time)
    
    restart_recording()


def export_csv_performance_report(csvfilename, dialect='excel', filename=None):
    '''
    Exports a performance report in comma-separated value format as the
    given filename.
    '''
    
    filename = filename or LOGFILE_NAME
    stats = pstats.Stats(filename)

    with open(csvfilename, 'wb') as f:
        csvwriter = csv.writer(f, dialect=dialect)
        csvwriter.writerow(['Function name', 'Number of calls', 'Total time', 'Cumulative time'])
        for (key, value) in stats.stats.items():
            (filename, count, func_name) = key
            (ccalls, ncalls, total_time, cumulative_time, callers) = value
            csvwriter.writerow(map(str, [func_name, ncalls, total_time, cumulative_time]))


def print_stats():
    '''
    Thin call to pstats-style printing
    '''
    
    global STOPPED
    
    # Need to stop recording for a moment to flush all the stats to disk
    if not STOPPED:
        stop_recording()
            
    stats = pstats.Stats(LOGFILE_NAME)
    stats.print_stats()
    
    restart_recording()

# At module exit, stop recording and save the file
import atexit
atexit.register(stop_recording)
