#   Copyright: Australian Bureau of Meteorology [2011]
#   Authors:
#       [Tennessee Leeuwenburg <t.leeuwenburg@bom.gov.au>]
#       [Edward Schofield <ed@pythoncharmers.com>]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

'''
This module exists to provide integration between the benchmarker.py module
and the codespeed benchmarking visualisation server. Benchmarker.py will
work perfectly well without this module, as its job is simply data collection.

This module is loosely coupled, which is to say it is an integrated system
but does not depend on any benchmarker.py code. It works by assuming a particular
directory structure containing pstats files.

/top-level-directory
    codespeed.config: contains the codespeed submission metadata
    /project_name (e.g. ...)
        /environment_name (e.g. "HP Z400")
            /executable_name (e.g. "pypy-c-jit")
                /YYYY
                   /MM
                      /DD
                         hostname.pstats
           
If revision-id tagging was enabled, then this will integrate with codespeed
and set revision ids with each submission. If not, the date and time will
be used to track the changes instead. 
'''

import os
import os.path
import sys
import pstats
from datetime import datetime
import urllib, urllib2
import ConfigParser

from configuration import DEFAULT_CODESPEED_URL, DEFAULT_METADATA_FILENAME, \
                          DEFAULT_STATS_LOCATION


def retrieve_executables(base_dir, project_name, environment_name):
    """
    Grabs the list of directories under base_dir/project_name/environment_name
    """
    # TODO: use a more efficient method for large trees
    path = os.path.join(base_dir, project_name, environment_name)
    assert os.path.exists(path)
    return list(os.walk(path))[0][1]


def load_metadata(base_dir=DEFAULT_STATS_LOCATION, filename=DEFAULT_METADATA_FILENAME):
    """
    Loads metadata (e.g. codespeed.config) from the given dir and filename.
    (Formatted in Windows INI style.) Returns a dictionary almost ready for
    submission to the codespeed server -- but with multiple projects, benchmark
    names, and executables, and without any actual results.
    """
    config = ConfigParser.ConfigParser()
    # Make the config parser case-sensitive:
    config.optionxform = str

    f = os.path.join(base_dir, filename)
    # config.read() doesn't raise an exception if it can't find the file, so we
    # catch it manually:
    if not os.path.exists(f):
        raise OSError('Cannot find metadata file ' + f)
    config.read(f)

    data = dict((key, config.get('Global', key)) \
                for key in ['branch', 'environment', 'commitid', 'codespeed_url'])
    data['projects'] = config.options('Benchmarks')
    data['benchmarks'] = dict((project, bm.split(', ')) \
                              for (project, bm) in config.items('Benchmarks'))

    # We take the benchmark names from the directories inside base_dir
    data['executables'] = dict([(project,
                                 retrieve_executables(base_dir, project, data['environment'])) \
                                 for project in data['projects']])

    return data


def date_from_path(pathname):
    """
    Returns a datetime object from the pathname where a pstats file is located.
    e.g. /base_dir/project/env/executable/2011/05/09 -> datetime(2011, 5, 9)
    """
    (year, month, day) = pathname[-10:].split(os.sep)
    return datetime(*map(int, (year, month, day)))


def load_results(metadata, base_dir=DEFAULT_STATS_LOCATION, ignore_broken_files=False):
    """
    Loads a tree full of pstats objects and returns a dict of dicts of the
    results (which are pstats.stats objects), one for each 4-tuple
    (project, environment, executable_name, benchmark_date).

    TODO: allow filtering by date?
    TODO: assign times correctly in the datetime objects. Currently these are
          0:00. The time should perhaps be stored in the pstats filename?
    """
    results = {}
    for project in metadata['projects']:
        # Currently the environment is fixed in the config file
        env = metadata['environment']
        for exe in metadata['executables'][project]:
            path = os.path.join(base_dir, project, env, exe)
            assert os.path.exists(path)
            # Now walk each dir tree looking for pstats files
            for (dirpath, dirnames, filenames) in os.walk(path):
                for f in filenames:
                    if f.endswith('.pstats'):
                        full_f = os.path.join(dirpath, f)
                        dt = date_from_path(dirpath)
                        # Load each pstats file
                        try:
                            stats = pstats.Stats(full_f)
                        except:
                            if not ignore_broken_files:
                                raise Exception("Could not load pstats file " + full_f)
                        results[project, env, exe, dt] = stats
    return results


def prepare_data(metadata, results, what='ct'):
    """
    Merges the metadata and the results dictionaries into a list of dicts, each
    ready for submitting to a codespeed server with send_one_result().

    The format of the stats dictionary inside the pstats object isn't
    explicitly documented, but some digging into pstats.py reveals it's:
        {function_name: (cc, nc, tt, ct, callers),
         ...
        }
    where cc is the number of primitive calls
          nc is the number of total calls?
          tt is the total time
          ct is the cumulative time

    The 'what' parameter is a string indicating one of these.
    """
    OUTPUT_OPTIONS = {'cc': 0, 'nc': 1, 'tt': 2, 'ct': 3}
    sendme = []
    for ((project, env, exe, dt), stats) in results.iteritems():
        data_template = {
            'commitid': metadata['commitid'],
            'branch': metadata['branch'],
            'environment': metadata['environment'],
            'project': project,
            'environment': env,
            'executable': exe,
            'result_date': str(dt)
        }
        benchmarks = set(metadata['benchmarks'][project])
        for (func, data) in stats.stats.iteritems():
            if func[2] in benchmarks:
                my_data = data_template.copy()
                # FIXME: the benchmark should be some combination of module and
                # function name
                my_data['benchmark'] = str(func[2])
                my_data['result_value'] = data[OUTPUT_OPTIONS[what]]
                sendme.append(my_data)

        # Other unused values: 'std_dev', 'max', 'min'

    return sendme


def send_one_result(data, codespeed_url=DEFAULT_CODESPEED_URL, verbose=True):
    """
    Submits a dict of data to the codespeed server with HTTP POST.
    """
    params = urllib.urlencode(data)
    if verbose:
        print "Executable %s, revision %s, benchmark %s" % (data['executable'], data['commitid'], data['benchmark'])
    url_part = codespeed_url + 'result/add/'
    if verbose:
        print url_part
    f = urllib2.urlopen(url_part, params)
    response = f.read()
    f.close()
    if verbose:
        print "Server (%s) response: %s\n" % (codespeed_url, response)


class UsageError(Exception):
    def __init__(self, msg):
        self.msg = msg


def main(argv=None):
    if argv is None:
        argv = sys.argv
    try:
        basedir = argv[1]
        assert isinstance(basedir, basestring)
    except:
        raise UsageError('Pass a single argument: the base directory for the pstats files')

    metadata = load_metadata(basedir)
    results = load_results(metadata, basedir)
    data = prepare_data(metadata, results)
    for datum in data:
        send_one_result(datum, metadata['codespeed_url'])


if __name__ == '__main__':
    main()

