#   Copyright [2011] [Tennessee Leeuwenburg <tleeuwenburg@gmail.com>]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import inspect
import os
import shutil
import unittest

import bench
from bench.benchmarker import benchmark

class TestBenchmarker(unittest.TestCase):

    def test_basic_benchmarking(self):
        '''
        An extremely basic benchmarking test
        '''
        
        @benchmark()
        def time_me():
            
            for _ in range(100):
                pass
            
        time_me()
        
    def test_args_and_kwargs(self):
        '''
        Test arguments are passed through correctly to the inner func
        '''
        
        @benchmark()
        def time_me(a, b="foo"):
            
            assert a is not None
            assert b == "foo"
            
        time_me("bar", b="foo")
        
    def test_multiple_calls(self):
        '''
        The the benchmarker can be called multiple times in the same
        benchmarking run
        '''
        
        @benchmark()
        def dummy_one():
            
            i = 0
            for _ in range(100):
                i + 5
                i = i / 5
                
        dummy_one()
                
        @benchmark()
        def dummy_two():
            
            import random
            i = 0.
            for _ in range(100):
                j = random.random()
                i += j
                
        dummy_two()
        
    def test_exceptions_passed_through(self):
        '''
        Benchmarking is run, even on functions which raise exceptions.
        The function is run N times, and the most-recent exception is
        re-raised afterwards.
        '''
        
        @benchmark(repeat=100, max_time_seconds=20)    
        def bar():
            
            for i in range(100):
                pass
            
            return 4. / 0 # cause div by zero error
            
        self.assertRaises(ZeroDivisionError, bar)
        
    def test_signatures_preserved(self):
        '''
        Normal decorators mash on the function signatures when wrapping 
        functions. This is a problem because sometimes code inspects the
        function signatures to determine behaviour. In my case, this was 
        done by py.test to determing the resources specified in the funcargs.
        
        So, I used the decorator module to munge on the function and produce
        a new function with the same signature as the decorated function ftw!
        '''
        
        def bar1(x):
            pass
        
        bar1_sig = inspect.getargspec(bar1)
        
        @benchmark()
        def bar2(x):
            pass
        
        bar2_sig = inspect.getargspec(bar2)
        
        assert bar1_sig == bar2_sig
        
    def test_creates_archive_directories_properly(self):        
        '''
        Create a bunch of fake nows, set the system time, create some pstats files, check
        it all works.
        '''
        
        tld = '/tmp/benchlog'
        project_name='tests'
        environment_name='testEnv'
        executable_name='tester'
        
        # Clean up before the test is run, so WARNING to any users not to
        # specify a tld which actually has non-test contents!
        if os.path.exists(tld):
            shutil.rmtree(tld)
        
        for i in range(5):
            bench.benchmarker.set_archiving(tld, 
                                            project_name,
                                            environment_name,
                                            executable_name)
            
            @benchmark()
            def time_me():
                
                for _ in range(100):
                    pass            
            
            time_me()
            
        bench.benchmarker.stop_recording()
        
        file_list = list(os.walk(tld))
        relevant_dir_listing = file_list[6][2]
        
        assert len(relevant_dir_listing) == 5

        # Clean up after the test is run also
        if os.path.exists(tld):
            shutil.rmtree(tld)        
                
    def test_print_stats(self):
        '''
        An extremely basic benchmarking test
        '''
        
        import bench
        import bench.benchmarker
        
        reload(bench)
        reload(bench.benchmarker)
        
        @benchmark()
        def time_me():
            
            for _ in range(100):
                pass
            
        time_me()
        bench.benchmarker.print_stats()
        
        time_me()
        bench.benchmarker.print_stats()
    
if __name__ == "__main__":   
    '''
    Just run some basic tests to see how to use the benchmarker suite
    '''
    
    unittest.main()
